"""
Conversions functions.

Library Version 1.0

Copyright 2013 Paul Griffiths
Email: mail@paulgriffiths.net

Distributed under the terms of the GNU General Public License.
http://www.gnu.org/licenses/
"""

# Set up constants for calculations. Centimetres,
# square centimetres, and cubic centimetres are the
# reference measures for length, area, and volume
# respectively.

# Metric lengths

_CM = 1
_MM = _CM / 10.0
_M = _CM * 100
_KM = _M * 1000
_NMI = _M * 1852

# Imperial lengths

_IN = _CM * 2.54
_PT = _IN / 72.0
_PI = _IN / 6.0
_FT = _IN * 12
_YD = _FT * 3
_FTM = _YD * 2
_FUR = _YD * 220
_MI = _FT * 5280
_LEA = _MI * 3

# Volumes

_CC = _CM ** 3
_L = _CC * 1000
_GAL = _CM * 4546.09
_CI = _IN ** 3
_USGAL = _CI * 231

# Weights

_G = 1
_KG = _G * 1000
_MG = _G / 1000.0
_LB = _KG / 2.2046213
_O = _LB / 16.0
_GRAIN = _O / 437.5

# Set up dictionaries of relative scales

_LENGTHS = {"point": _PT,
            "points": _PT,
            "pt": _PT,
            "pts": _PT,
            "pica": _PI,
            "picas": _PI,
            "pi": _PI,
            "mm": _MM,
            "millimetre": _MM,
            "millimetres": _MM,
            "millimeter": _MM,
            "millimeters": _MM,
            "cm": _CM,
            "centimetre": _CM,
            "centimetres": _CM,
            "centimeter": _CM,
            "centimeters": _CM,
            "inch": _IN,
            "inches": _IN,
            "in": _IN,
            "foot": _FT,
            "feet": _FT,
            "ft": _FT,
            "yard": _YD,
            "yards": _YD,
            "yd": _YD,
            "m": _M,
            "metre": _M,
            "metres": _M,
            "meter": _M,
            "meters": _M,
            "fathom": _FTM,
            "fathoms": _FTM,
            "ftm": _FTM,
            "furlong": _FUR,
            "furlongs": _FUR,
            "fur": _FUR,
            "km": _KM,
            "kilometre": _KM,
            "kilometres": _KM,
            "kilometer": _KM,
            "kilometers": _KM,
            "mile": _MI,
            "miles": _MI,
            "mi": _MI,
            "nm": _NMI,
            "nmi": _NMI,
            "nautmile": _NMI,
            "nautmiles": _NMI,
            "nauticalmile": _NMI,
            "nauticalmiles": _NMI,
            "league": _LEA,
            "leagues": _LEA,
            "lea": _LEA}

_AREAS = {"point2": _PT ** 2,
          "points2": _PT ** 2,
          "pt2": _PT ** 2,
          "pts2": _PT ** 2,
          "pica2": _PI ** 2,
          "picas2": _PI ** 2,
          "pi2": _PI ** 2,
          "mm2": _MM ** 2,
          "millimetre2": _MM ** 2,
          "millimetres2": _MM ** 2,
          "millimeter2": _MM ** 2,
          "millimeters2": _MM ** 2,
          "cm2": _CM ** 2,
          "centimetre2": _CM ** 2,
          "centimetres2": _CM ** 2,
          "centimeter2": _CM ** 2,
          "centimeters2": _CM ** 2,
          "inch2": _IN ** 2,
          "inches2": _IN ** 2,
          "in2": _IN ** 2,
          "foot2": _FT ** 2,
          "feet2": _FT ** 2,
          "ft2": _FT ** 2,
          "yard2": _YD ** 2,
          "yards2": _YD ** 2,
          "yd2": _YD ** 2,
          "m2": _M ** 2,
          "metre2": _M ** 2,
          "metres2": _M ** 2,
          "meter2": _M ** 2,
          "meters2": _M ** 2,
          "centiare": _M ** 2,
          "ca": _M ** 2,
          "fathom2": _FTM ** 2,
          "fathoms2": _FTM ** 2,
          "ftm2": _FTM ** 2,
          "acre": _FT ** 2 * 43560,
          "acres": _FT ** 2 * 43560,
          "ac": _FT ** 2 * 43560,
          "are": _M ** 2 * 100,
          "ares": _M ** 2 * 100,
          "a": _M ** 2 * 100,
          "decare": _M ** 2 * 1000,
          "decares": _M ** 2 * 1000,
          "daa": _M ** 2 * 1000,
          "hectare": _M ** 2 * 10000,
          "hectares": _M ** 2 * 10000,
          "ha": _M ** 2 * 10000,
          "furlong2": _FUR ** 2,
          "furlongs2": _FUR ** 2,
          "fur2": _FUR ** 2,
          "km2": _KM ** 2,
          "kilometre2": _KM ** 2,
          "kilometres2": _KM ** 2,
          "kilometer2": _KM ** 2,
          "kilometers2": _KM ** 2,
          "mile2": _MI ** 2,
          "miles2": _MI ** 2,
          "mi2": _MI ** 2,
          "nmi2": _NMI ** 2,
          "nautmile2": _NMI ** 2,
          "nautmiles2": _NMI ** 2,
          "nauticalmile2": _NMI ** 2,
          "nauticalmiles2": _NMI ** 2,
          "league2": _LEA ** 2,
          "leagues2": _LEA ** 2,
          "lea2": _LEA ** 2}

_VOLUMES = {"point3": _PT ** 3,
            "points3": _PT ** 3,
            "pt3": _PT ** 3,
            "pts3": _PT ** 3,
            "pica3": _PI ** 3,
            "picas3": _PI ** 3,
            "pi3": _PI ** 3,
            "mm3": _MM ** 3,
            "millimetre3": _MM ** 3,
            "millimetres3": _MM ** 3,
            "millimeter3": _MM ** 3,
            "millimeters3": _MM ** 3,
            "ml": _CC,
            "millilitre": _CC,
            "millilitres": _CC,
            "milliliter": _CC,
            "milliliters": _CC,
            "cm3": _CC,
            "cc": _CC,
            "centimetre3": _CC,
            "centimetres3": _CC,
            "centimeter3": _CC,
            "centimeters3": _CC,
            "inch3": _CI,
            "inches3": _CI,
            "in3": _CI,
            "uspint": _USGAL / 8.0,
            "uspints": _USGAL / 8.0,
            "uspt": _USGAL / 8.0,
            "uspts": _USGAL / 8.0,
            "usquart": _USGAL / 4.0,
            "usquarts": _USGAL / 4.0,
            "usqt": _USGAL / 4.0,
            "usqts": _USGAL / 4.0,
            "usgallon": _USGAL,
            "usgallons": _USGAL,
            "usgal": _USGAL,
            "usgals": _USGAL,
            "pint": _GAL / 8.0,
            "pints": _GAL / 8.0,
            "pt": _GAL / 8.0,
            "pts": _GAL / 8.0,
            "quart": _GAL / 4.0,
            "quarts": _GAL / 4.0,
            "qt": _GAL / 4.0,
            "qts": _GAL / 4.0,
            "gallon": _GAL,
            "gallons": _GAL,
            "gal": _GAL,
            "gals": _GAL,
            "l": _L,
            "litre": _L,
            "litres": _L,
            "liter": _L,
            "liters": _L,
            "foot3": _FT ** 3,
            "feet3": _FT ** 3,
            "ft3": _FT ** 3,
            "yard3": _YD ** 3,
            "yards3": _YD ** 3,
            "yd3": _YD ** 3,
            "m3": _M ** 3,
            "metre3": _M ** 3,
            "metres3": _M ** 3,
            "meter3": _M ** 3,
            "meters3": _M ** 3,
            "fathom3": _FTM ** 3,
            "fathoms3": _FTM ** 3,
            "ftm3": _FTM ** 3,
            "furlong3": _FUR ** 3,
            "furlongs3": _FUR ** 3,
            "fur3": _FUR ** 3,
            "km3": _KM ** 3,
            "kilometre3": _KM ** 3,
            "kilometres3": _KM ** 3,
            "kilometer3": _KM ** 3,
            "kilometers3": _KM ** 3,
            "mile3": _MI ** 3,
            "miles3": _MI ** 3,
            "mi3": _MI ** 3,
            "nmi3": _NMI ** 3,
            "nautmile3": _NMI ** 3,
            "nautmiles3": _NMI ** 3,
            "nauticalmile3": _NMI ** 3,
            "nauticalmiles3": _NMI ** 3,
            "league3": _LEA ** 3,
            "leagues3": _LEA ** 3,
            "lea3": _LEA ** 3}

_WEIGHTS = {"gram": _G,
            "grams": _G,
            "g": _G,
            "decigram": _G / 10.0,
            "decigrams": _G / 10.0,
            "dg": _G / 10.0,
            "centigram": _G / 100.0,
            "centigrams": _G / 100.0,
            "cg": _G / 100.0,
            "milligram": _MG,
            "milligrams": _MG,
            "mg": _MG,
            "decagram": _G * 10,
            "decagrams": _G * 10,
            "dag": _G * 10,
            "hectogram": _G * 100,
            "hectograms": _G * 100,
            "hg": _G * 100,
            "kilogram": _KG,
            "kilograms": _KG,
            "kg": _KG,
            "tonne": _KG * 1000,
            "megagram": _KG * 1000,
            "t": _KG * 1000,
            "grain": _GRAIN,
            "grains": _GRAIN,
            "gr": _GRAIN,
            "dram": _O / 16.0,
            "drams": _O / 16.0,
            "dr": _O / 16.0,
            "ounce": _O,
            "ounces": _O,
            "oz": _O,
            "pound": _LB,
            "pounds": _LB,
            "lb": _LB,
            "lbs": _LB,
            "hundredweight": _LB * 100,
            "cwt": _LB * 100,
            "ton": _LB * 2000}

# Set up collection of quantities for conversion functions

_COLLECTIONS = [_LENGTHS, _AREAS, _VOLUMES, _WEIGHTS]


# Define custom exceptions

class InvalidUnit(Exception):

    """
    Exception for unrecognized unit.
    """

    pass


class InvalidLengthUnit(InvalidUnit):

    """
    Exception for unrecognized length unit.
    """

    pass


class InvalidAreaUnit(InvalidUnit):

    """
    Exception for unrecognized area unit.
    """

    pass


class InvalidVolumeUnit(InvalidUnit):

    """
    Exception for unrecognized volume unit.
    """

    pass


class InvalidWeightUnit(InvalidUnit):

    """
    Exception for unrecognized weight unit.
    """

    pass


class IncompatibleUnits(Exception):

    """
    Exception for incompatible units (e.g. comparing miles to gallons).
    """

    pass


# Define functions

def validate_length(unit):

    """
    Checks if a length unit is recognized.

    Arguments:
    unit -- string containing the length unit to check
    """

    if not unit in _LENGTHS:
        raise InvalidLengthUnit(unit)


def validate_area(unit):

    """
    Checks if an area unit is recognized.

    Arguments:
    unit -- string containing the area unit to check
    """

    if not unit in _AREAS:
        raise InvalidAreaUnit(unit)


def validate_volume(unit):

    """
    Checks if a volume unit is recognized.

    Arguments:
    unit -- string containing the volume unit to check
    """

    if not unit in _VOLUMES:
        raise InvalidVolumeUnit(unit)


def validate_weight(unit):

    """
    Checks if a weight unit is recognized.

    Arguments:
    unit -- string containing the weight unit to check
    """

    if not unit in _WEIGHTS:
        raise InvalidWeightUnit(unit)


def validate_unit(unit):

    """
    Checks if a unit is recognized.

    Arguments:
    unit -- string containing the unit to check
    """

    for coll in _COLLECTIONS:
        if unit in coll:
            return

    raise InvalidUnit(unit)


def validate_compatible_units(unit1, unit2):

    """
    Checks if two units are of a consistent type.

    Arguments:
    unit1, unit2 -- strings containing the units to check
    """

    # Check whether both units are valid, first...

    for unit in [unit1, unit2]:
        validate_unit(unit)

    # ...then check whether they're compatible.

    for coll in _COLLECTIONS:
        if unit1 in coll and unit2 in coll:
            return

    raise IncompatibleUnits("{0} and {1}".format(unit1, unit2))


def conversion_factor(unitfrom, unitto):

    """
    Returns a conversion factor for two units.

    Arguments:
    unitfrom -- string containing the unit from which to convert
    unitto -- string containing the unit to which to convert
    """

    unitfrom = unitfrom.lower()     # To make case-insensitive
    unitto = unitto.lower()         # To make case-insensitive

    validate_compatible_units(unitfrom, unitto)

    for coll in _COLLECTIONS:
        if unitfrom in coll and unitto in coll:
            return coll[unitfrom] / float(coll[unitto])


def convert(quantity, unitfrom, unitto):

    """
    Returns a converted quantity.

    Arguments:
    quantity -- quantity to convert, in 'unitfrom' units
    unitfrom -- string containing the unit from which to convert
    unitto -- string containing the unit to which to convert
    """

    return quantity * conversion_factor(unitfrom, unitto)
