"""
Materials classes.

Library Version 1.0

Copyright 2013 Paul Griffiths
Email: mail@paulgriffiths.net

Distributed under the terms of the GNU General Public License.
http://www.gnu.org/licenses/
"""


from geometry.conversions import convert


class _Material():

    """
    Base material class
    """

    def __init__(self):
        self._specific_gravity = 1

    def weight(self, quantity=1, volumeunits="cc", weightunits="grams"):

        """
        Returns the weight of a specified volume of the material.

        Arguments:
        quantity -- the volume
        volumeunits -- units in which 'quantity' is stated
        weightunits -- units for which to return the weight

        For example, weight(35, "in3", "pounds") will return the weight,
        in pounds, of 35 cubic inches of the material.
        """

        cc_quant = convert(quantity, volumeunits, "cc")
        w_grams = cc_quant * self._specific_gravity
        return convert(w_grams, "grams", weightunits)

    def specific_gravity(self):

        """
        Returns the specific gravity of the material.
        """

        return self._specific_gravity


class DefaultMaterial(_Material):

    """
    Default material class
    """

    def __init__(self):
        _Material.__init__(self)
        self._specific_gravity = 1


class Water(_Material):

    """
    Water material class
    """

    def __init__(self):
        _Material.__init__(self)
        self._specific_gravity = 1


class _Metal(_Material):

    """
    Generic metal class
    """

    def __init__(self):
        _Material.__init__(self)


class MildSteel(_Metal):

    """
    Steel material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 7.861093


class Aluminium(_Metal):

    """
    Aluminium material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 2.6989


class Copper(_Metal):

    """
    Copper material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 8.96


class Gold(_Metal):

    """
    Gold material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 19.32


class Silver(_Metal):

    """
    Silver material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 10.5


class Iron(_Metal):

    """
    Iron material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 7.894


class Lead(_Metal):

    """
    Lead material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 11.35


class Tin(_Metal):

    """
    Tin material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 7.31


class Zinc(_Metal):

    """
    Zinc material class
    """

    def __init__(self):
        _Metal.__init__(self)
        self._specific_gravity = 7.135
